// 
// Module: CorePosData
// Author: Richard Saxton
//
// Description: Contains all the code to extract and manipulate
//              positions and attitude information
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
/* General */

#include "Xmm.h"
#include "common.h"
#include "OalServer.h"
#include "CalServer.h"
#include "DataSubspace.h"
#include "DssBuilder.h"
#include "Tracer.h"
#include "CcdQuantumDataServer.h"
#include "ExceptionErrorHandler.h"

/* Local classes */
#include "CorePosData.h"
#include "RegionExtent.h"
#include "PixCoords.h"

/* Constructors */

CorePosData::CorePosData( DataSet * spectrumset , Table * spectrumtable)
{
  PROC; 

/*  Read the essential information from the header */

  string datamode = spectrumset->attribute("DATAMODE")-> asString();
  string dateobs = spectrumset->attribute("DATE-OBS")->asString();
  string submode = spectrumset->attribute("SUBMODE")->asString();

  /* Set window mode - need something to avoid crash out here for
  mode='Unknown' */

  ExceptionErrorHandler local;
  ErrorHandler* oldHandler = errHandler.handler(&local);
  Xmm::Mode windowMode;
  try {
    windowMode = Xmm::toMode(submode);
  } catch(SasException::Error) { errstr<<"Submode not valid - using FullWindow"
<<Sas::warning("UnknownModeString");
        windowMode = Xmm::PRIME_FULL_WINDOW; // d/f
  }
  errHandler.handler(oldHandler);

  calServer.state()->mode()->set(windowMode);

  /* Get pointing info */
  double rapnt = spectrumset->
    attribute("RA_PNT")->
    asReal();

  double decpnt = spectrumset->
    attribute("DEC_PNT")->
    asReal();

  double papnt = spectrumset->
    attribute("PA_PNT")->
    asReal();

  /* Read info from the header of the spectrum extension */
  _dss = new DataSubspace(spectrumtable);

  _instrumentID =  Xmm::toInstrument(spectrumtable->
                       attribute("INSTRUME")->asString());

  /* Set up a coordinate conversion object */
  _poswcs = new PosWCS(spectrumset, rapnt, decpnt, papnt, false);

  /* Create a position utilities object */
  _positionUtils = new PositionUtils(_instrumentID);


    errstr
    << "Set up poswcs and positionUtils"
    << Sas::message(Msg::App, Msg::Verbose);


  /* For timing and burst modes the spatial selection is either left blank
     or is made in RAW-X. In this case a pseudo-region should be generated
     in DET coords to handle vignetting and PSF issues */

  if (windowMode == Xmm::FAST_TIMING || windowMode == Xmm::FAST_BURST
                                     || windowMode == Xmm::FAST_UNCOMPRESSED
                                     || windowMode == Xmm::FAST_COMPRESSED) {

     addDetRegion(_positionUtils);
  }


  /* Create a source box extent object */
  RegionExtent * extent = new RegionExtent(_dss, _positionUtils, 
                                             datamode, dateobs, _poswcs);

  /* Get the source box centre detector coordinates */
  _sourcedetx = (extent->xmin() + extent->xmax()) / 2.0;
  _sourcedety = (extent->ymin() + extent->ymax()) / 2.0;

     errstr 
       << "source position: "<<_sourcedetx<<" "<<_sourcedety
       << Sas::message(Msg::App, Msg::Verbose);

ENDPROC;
};

/*  Destructors  */
CorePosData::~CorePosData()
{
  PROC;

  delete _positionUtils;   _positionUtils  = 0;
  delete _poswcs;   _poswcs = 0;
  delete _dss;   _dss = 0;

  ENDPROC;
};

/* Return the source centre in raw coordinates */
PixCoords CorePosData::centre_raw() const
{
  
  PixCoords pixCoord(0,0);

  errstr 
       << "source position: "<<_sourcedetx<<" "<<_sourcedety
       << Sas::message(Msg::App, Msg::Verbose);

  /* Check if the source centre lies on a chip */
  CamCoords testpixel = _positionUtils->detxy005ToCamCoord2(
                                           _sourcedetx,_sourcedety);

  GeometryDataServer * ct = 0; ct = calServer.getAtom(ct);

  if ( ct->onCcd(testpixel.x(), testpixel.y()) == Xmm::CCD_NONE ) 
  {
   /* Source is off-chip so set the raw coordinates as 32.0, the Y
      value calculated from the DET-Y position */

     double rawX = 32.0;
     double rawY = 199.0 - abs(_sourcedety + 1000) * 199.0 / 16000.0;
     rawY = rawY < 1 ? 1:rawY;

     pixCoord = PixCoords(UNT(rawX), UNT(rawY));

     errstr 
       << "source is not centred on a CCD - generating RMF for "
       << "line number "<< rawY
       << Sas::message(Msg::App, Msg::Verbose);
  }
  else 
  {
 
     pixCoord = _positionUtils->camCoord2ToPixCoord(
            _positionUtils->detxy005ToCamCoord2(_sourcedetx,_sourcedety));
  }

  
  RETURN  pixCoord;
}

/* Get selected patterns from DSS */
EventGrades CorePosData::patternCode() const
{
  PROC;

  EventGrades pcode=CombinedEventGrades::SDTQ;
  double lowest=12, highest=0;
  double upper, lower;
  RangeBound upperB, lowerB;

  list <Range<double> > ranges = _dss->ranges("PATTERN");

  if (ranges.size()==0)
  {
      errstr 
	<< "There is no PATTERN range info in the DSS. Assume the full range of patterns has been used."
	<< Sas::warning("NoDSSPatterns");

     /* Default to singles, doubles, triples and quads for MOS and
        singles + doubles for PN*/
      if (_instrumentID == Xmm::EPN)
      {
        pcode=CombinedEventGrades::SD;
      }
      else {
        pcode=CombinedEventGrades::SDTQ;
      }
  
  } else {

   /* Find the lowest and highest pattern values wanted */
    for ( list<Range<double> >::iterator i = ranges.begin(); 
	i != ranges.end(); 
	++i )
     {
      (*i).values(lower, upper, lowerB, upperB);

      if (lowerB==Unbounded)
	{
	  lower = 0;
	}
      else if (lowerB==Open)
        {
          ++lower;
        }

      if (upperB==Unbounded)
	{
	  upper = 12;
	}
      else if (upperB==Open)
        {
          --upper;
        }
	  
      lowest = lower < lowest ? lower:lowest;
      highest = upper > highest ? upper:highest;
     }
  
     errstr 
       << "Pattern range in file: "<<lowest<<":"<<highest
       << Sas::message(Msg::App, Msg::Verbose);

     if (lowest==0 && highest==0)
       {
        //pcode=PatternFractionDataServer::SINGLE;
        pcode=CombinedEventGrades::SINGLE;
       }
     else if (lowest==0 && highest==4)
       {
        pcode=CombinedEventGrades::SD;
       }
     else if (lowest==0 && highest==12)
       {
        pcode=CombinedEventGrades::SDTQ;
       }
     else if (lowest==1 && highest==4)
       {
        pcode=CombinedEventGrades::DOUBLE;
       }
     else if (lowest==5 && highest==8)
       {
        pcode=CombinedEventGrades::TRIPLE;
       }
     else if (lowest==9 && highest==12)
       {
        pcode=CombinedEventGrades::QUADRUPLE;
       }
     else
       {
          errstr
        << "An RMF cannot be produced for the pattern range in the DSS"
        << Sas::error("InvalidPatterns");
       }
   }

  /* need to put a temporary fiddle in here for MOS Timing mode.
     MOS Timing mode spectra are produced from pattern 0 events but
     these actually combine image mode pattern 0,1 and 3 events.
     This should be handled within the EMOSn_REDIST CCF file but
     currently isn't so here we will fix the response to be Singles+dubs */

   if ( (_instrumentID == Xmm::EMOS1 || _instrumentID == Xmm::EMOS2) && 
          calServer.state()->mode()->get() == Xmm::FAST_UNCOMPRESSED) {
      pcode=CombinedEventGrades::SD;
   }

   errstr << "Pattern code="<<pcode
          << Sas::message(Msg::App, Msg::Verbose);
       

   RETURN pcode;

  ENDPROC;
}

/* addDetRegion - adds a pseudo spatial region in detector coords to the dss */
void CorePosData::addDetRegion(PositionUtils * putils)
{
  PROC;

  /* For timing and burst modes the detector position is unimportant
     so dummy up a position in detector coords in the centre of the
     chip */

    double lower_rawx=30, lower_rawy=100;
    double upper_rawx=50, upper_rawy=120;

   /* Now find the detector coords of this position */
    PixCoords pixCoord1(UNT(lower_rawx), UNT(lower_rawy));
    CamCoords camCoord2 = putils->PixCoordTocamCoord2(pixCoord1);
    CoordPair detlow = putils->CamCoord2Todetxy005(camCoord2);
    double det_xlow = detlow.x();
    double det_ylow = detlow.y();
    PixCoords pixCoord2(UNT(upper_rawx), UNT(upper_rawy));
    camCoord2 = putils->PixCoordTocamCoord2(pixCoord2);
    CoordPair detupp = putils->CamCoord2Todetxy005(camCoord2);
    double det_xupp = detupp.x();
    double det_yupp = detupp.y();

    double detxcen = (det_xlow + det_xupp) / 2.0;
    double detycen = (det_ylow + det_yupp) / 2.0;
    double detx_hwid = abs(detxcen - det_xlow);
    double dety_hwid = abs(detycen - det_ylow);

    errstr<<"Det pixel centre and width "<<detxcen<<" "<<detycen
              <<detx_hwid<<" "<<dety_hwid
              << Sas::message(Msg::App, Msg::Noisy);

   /* Create an expression from these coords */
    std::string expression;
   // expression="(DETX,DETY) in BOX("+toString(detxcen)+","+toString(detycen)+
    //           ","+toString(detx_hwid)+","+toString(dety_hwid);
    expression="(DETX,DETY) in BOX(100,100,100,100";
    expression += ",0)";

    errstr<<"From raw X: "<<lower_rawx<<" "<<upper_rawx
              <<" raw Y: "<<lower_rawy<<" "<<upper_rawy<<" got: "
              <<expression
              << Sas::message(Msg::App, Msg::Verbose);

   /* Create a filter */
    DataSubspace _dssextra = dssBuilder.getDss(expression);

    for(unsigned icomp=0; icomp <_dssextra.components(); icomp++){

      errstr<<"Adding component "<<icomp<<" to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

      _dss->addComponent(*_dssextra.component(icomp));
    }

    errstr<<"Added new component to DSS"
              << Sas::message(Msg::App, Msg::Verbose);

 ENDPROC;
}

