// Module: FixedMatrixColumn
// Author: Hassan Siddiqui
//
// Description:
// Fixed length vector column version of MatrixColumn
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: FixedMatrixColumn.cc,v 1.2 2000/03/09 13:37:36 hsiddiqu Exp $

/* General */

/* SAS-specific */
#include "Dal.h"
#include "errstr.h"
#include "Tracer.h"

/* Local */
#include "FixedMatrixColumn.h"

/* Constructor */
FixedMatrixColumn::FixedMatrixColumn( Table * table,
				      const vector<unsigned long>& sizeVect):
  _matrixCol(0),
  _matrixColData(0),
  _matrixData(0),
  _row(0),
  _largestelement(0)
{
  PROC;

  errstr << " creating FixedMatrixColumn object ... "
	 << Sas::message(Msg::App, Msg::Verbose);

  /* Check that sizeVect is scalar */
  if (sizeVect.size() != 1)
    {
      errstr << " sizeVect has " 
	     << sizeVect.size() << " dimensions."
	     << " expected 1. "
	     << Sas::error("matrixDimensionsInvalid");
      ALTRETURN;
    }

  /* Create new Column */
  _matrixCol = table->addColumn("MATRIX", 
				Column::Real32, 
				"No comment", 
				"", 
				sizeVect);

  _rowelements = sizeVect[0];

  /* Get matrixColData and matrixData objects */
  _matrixColData = _matrixCol->data();
  _matrixData    = _matrixColData  ->real32Data();

  /* Initialise data */
  errstr << " Initialising MATRIX column with zeroes ..."
	 << Sas::message(Msg::App, Msg::Verbose);

  //  unsigned int nelements = sizeVect[0];

//    for (unsigned int row = 0; row < table->rows(); ++row)
//      {
//        errstr << " Filling row " << row
//  	     << Sas::message(Msg::App, Msg::Verbose);

//        for (unsigned int element = 0; element < nelements; ++element)
//  	{
//  	  _matrixData[_matrixColData->position( row, element)] = 0;
//    	}

//        errstr << " Filled row  " << row << " ok."
//  	     << Sas::message(Msg::App, Msg::Verbose);
//      }

  errstr << " FixedMatrixColumn object created. "
	 << Sas::message(Msg::App, Msg::Verbose);

  ENDPROC;
}


/* Destructor */
FixedMatrixColumn::~FixedMatrixColumn()
{
  PROC;
  
  errstr << " Destroying FixedMatrixColumn object ... "
	 << Sas::message(Msg::App, Msg::Verbose);
 
 if (!_matrixColData)
    {
      errstr << " matrixColData object for FixedMatrixColumn is not set in destructor. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixColData");
      ALTRETURN;
    }
  
  delete _matrixColData; _matrixColData = 0;

  ENDPROC;
}

/* Set state to a particular row, and the size of that row */
void FixedMatrixColumn::row( unsigned long row,
			     unsigned int /* elements NOT USED */ )
{
  _row = row;
}

/* Write to a given position in column */
void FixedMatrixColumn::writeTo( unsigned int element,
				 double value)
{
  PROC;
  
  if ( element > _rowelements)
    {
      errstr << " element specified [" << element
	     << "] is out of bounds for matrix column [0:" << _rowelements-1
	     << Sas::error("elementOutOfBounds");
      ALTRETURN;
    }

  if (!_matrixColData)
    {
      errstr << " matrixColData object for VarMatrixColumn is not set in writeTo() method. This should not occur, please raise an SPR."
	     << Sas::error("invalidMatrixColData");
      ALTRETURN;
    }

  _matrixData[_matrixColData->position( _row, element)] = value;

  errstr << "matrix ["
	 << _row
	 << ","
	 << element
	 << "] set to "
	 << _matrixData[_matrixColData->position(_row, element)]
	 << "( expected " << value << " )"
	 << Sas::message(Msg::App, Msg::Verbose);

  /* Set _largestelement if appropriate */
  if (element > _largestelement)
    {
      _largestelement = element;
    }

  ENDPROC;
}

unsigned int FixedMatrixColumn::rowElements() const
{
  PROC;
  RETURN _rowelements;
}


void FixedMatrixColumn::closeRow()
{
  PROC;

  if (_largestelement>=_rowelements-1)
    {
      ALTRETURN;
    }

  /* Pad remainder of row with zeroes */
  errstr << " Padding row " << _row
	 << " elements " << _largestelement+1
	 << " to " << _rowelements-1
	 << " with zeroes..."
	 << Sas::message(Msg::App, Msg::Verbose);

  for (unsigned int element = _largestelement+1; 
       element <_rowelements;
       ++element)
    {
	  _matrixData[_matrixColData->position( _row, element)] = 0;
    }
  ENDPROC;
}
