// Module: Rmfgen
// Author: Hassan Siddiqui
//
// Description: Main class for generating RMF
// 
// Comments: 
//  
// Modifications: See ChangeLog
// 
// $Id: Rmfgen.cc,v 1.12 2000/11/08 15:46:59 hsiddiqu Exp $

// General classes
#include <iostream>

// SAS-specific classes
#include "param.h"
#include "Dal.h"
#include "FileLocator.h"
#include "Tracer.h"
#include "CalServer.h"
#include "CcdQuantumDataServer.h"
#include "Proxy.h"
#include "Ccf.h"

// Local classes
#include "Rmfgen.h"
#include "ConcreteRmfDataServer.h"
#include "RmfTableIterator.h"
#include "ResponseGrouping.h"
#include "CorePosData.h"

void Rmfgen::exec()
{
  
  PROC;

  /* Get spectrum info */
  string spectrumsetname=stringParameter("spectrumset");
  DataSet * specSet = dataSetServer->open(spectrumsetname, 
					  DataSet::Read);

  Table * specTable = specSet->table("SPECTRUM");

  Column * specchannelCol = specTable->column("CHANNEL");

  unsigned tlmin  = specchannelCol->attribute("TLMIN")->asInt();
  
  if (tlmin != 0)
    {
      errstr << "TLMIN for column CHANNEL has an invalid value ["
	     << tlmin 
	     << "] it must be zero "
	     << Sas::error("invalidChannelLowBound");

      ALTRETURN;
    }

  unsigned tlmax  = specchannelCol->attribute("TLMAX")->asInt();
  _detchans = specTable->attribute("DETCHANS")->asInt();
  
  if (_detchans > tlmax-tlmin+1)
    {
      errstr 
	<< " From dataset " << spectrumsetname
	<< " DETCHANS[" << _detchans
	<< "], TLMIN[" << tlmin 
	<< "], and TLMAX[" << tlmax 
	<< "] are incompatible "
	<< Sas::error("invalidChannelBounds");

      ALTRETURN;
    }

    _telescope = specSet->attribute("TELESCOP")->asString();
    errstr << " The telescope is " << _telescope
  	 << Sas::message(Msg::App,Msg::Verbose);

    _instrument = specSet->attribute("INSTRUME")->asString();
    errstr << " The instrument is " << _instrument
  	 << Sas::message(Msg::App,Msg::Verbose);

    _filter = specSet->attribute("FILTER")->asString();
    errstr << " The filter is " << _filter
  	 << Sas::message(Msg::App,Msg::Verbose);

    _submode = specSet->attribute("SUBMODE")->asString();
    errstr << " The windowing mode is " << _submode
  	 << Sas::message(Msg::App,Msg::Verbose);

    unsigned specpix  = 0;
    double   specval  = 0;
    unsigned specdelt = 1;

   /* Set the instrument and CCD states in the calServer */
    calServer.setInstrument(Xmm::toInstrument(_instrument));
  
    if (Xmm::toInstrument(_instrument) == Xmm::EPN) {
      calServer.state()->ccd()->set(EPNcam::CCD_1);
    } else {   // MOS
      calServer.state()->ccd()->set(EMOScam::CCD_1);
    }

   /* Set up a position and dss processing object */
    CorePosData * corePosData = new CorePosData(specSet, specTable);

       errstr << "Core pos gives: "<<corePosData->centre_raw().x()
         <<" "<<corePosData->centre_raw().y()
         << Sas::message(Msg::App,Msg::Verbose);

    /* Get attributes describing rebinning */
    if (specTable->hasAttribute("SPECPIX"))
      {
	specpix  = specTable->attribute("SPECPIX")->asInt();
      }
    else
      {
	errstr << "Attribute SPECPIX is not defined in " 
	       << spectrumsetname 
	       << ":SPECTRUM . Assuming value to be " << specpix
	       << Sas::warning("specpixNotDefined");
      }

    if (specTable->hasAttribute("SPECDELT"))
      {
	specdelt  = specTable->attribute("SPECDELT")->asInt();
      }
    else
      {
	errstr << "Attribute SPECDELT is not defined in " 
	       << spectrumsetname 
	       << ":SPECTRUM . Assuming value to be " << specdelt
	       << Sas::warning("specdeltNotDefined");
      }

    if (specTable->hasAttribute("SPECVAL"))
      {
	specval  = specTable->attribute("SPECVAL")->asReal();
      }
    else
      {
	errstr << "Attribute SPECVAL is not defined in " 
	       << spectrumsetname 
	       << ":SPECTRUM . Assuming value to be " << specval
	       << Sas::warning("specvalNotDefined");
      }

    /* Old formula  - keep in case */
//    double   offsetD   = 
//      double(specval) + 0.5 - 
//      ((double(specpix-tlmin)+0.5)*double(specdelt));

    double  offsetD = specval - (specpix*specdelt) - 0.5*(specdelt-1);

    unsigned offsetI   = int(offsetD);
    if (double(offsetI) != offsetD)
    {
      errstr << "From SPECPIX=" << specpix
	     << ", SPECDELT=" << specdelt
	     << ", SPECVAL=" << specval
	     << ", TLMIN="  <<tlmin
	     << " got an offset of " << offsetD
	     << " which is not a round number."
	     << Sas::error("invalidChannelOffset");

      ALTRETURN;
    }
  errstr << "From SPECPIX=" << specpix
	 << ", SPECDELT=" << specdelt
	 << ", SPECVAL=" << specval
	 << ", TLMIN="  <<tlmin
	 << " got an offset of " << offsetD
	 << Sas::message(Msg::App, Msg::Verbose);

  dataSetServer->close(specSet);

  /* Check that the spectrum isn't too finely binned */ 
  /*
  if (Xmm::toInstrument(_instrument) == Xmm::EPN) {
     if (_detchans > 4096) {
        errstr<<"Spectrum contains too many bins, maximum for PN is 4096"
	      << Sas::error("TooManyChannels");
     }
  }
  else {
     if (_detchans > 800) {
        errstr<<"Spectrum contains too many bins, maximum for MOS is 800"
              << Sas::error("TooManyChannels");
     }
  }
  */

  /* If the instrument is PN get the line number from the header */
  unsigned pnline=0;

  if (Xmm::toInstrument(_instrument) == Xmm::EPN) {
     pnline = corePosData->centre_raw().y(); 

     errstr << " Using a PN line number: "<<pnline
	 << Sas::message(Msg::App,Msg::Verbose);
  }
  
  // Create rmf configuration object
  errstr << " Creating RmfDataServer..."
	 << Sas::message(Msg::App,Msg::Sparse);

  RmfDataServer * rmfDataServer = 
    new ConcreteRmfDataServer( _instrument,
			       tlmin,
			       tlmax,
			       _detchans,
			       specdelt,
			       offsetI,
			       realParameter   ("energymin"),
			       realParameter   ("energymax"),
			       intParameter    ("nenergybins"),
			       booleanParameter("withenergybins"),
			       corePosData->centre_raw(),
                               _submode,
			       stringParameter ("format"),
                               corePosData->patternCode() );

  errstr << " RmfDataServer created ok."
	 << Sas::message(Msg::App,Msg::Sparse);

  // Check constraints on i/p spectrum and rmf

  unsigned firstchannel = rmfDataServer->response()->firstChannel();
  if (tlmin != firstchannel)
    {
      errstr 
	<< " lower allowed channel bound of spectrum [" << tlmin
	<< "] differs from that used for the RMF [" << firstchannel
	<< "]. These datasets may not be compatible."
	<< Sas::warning("incompatibleLowerChannelBounds");
    }

  // Get the number of detector channels
  if (_detchans != rmfDataServer->response()->numberOfChannels())
    {
      errstr << "Rmfgen: number of channels calculated ["
	     << rmfDataServer->response()->numberOfChannels()
	     <<"] differs from that in input spectrum dataset ["
	     << _detchans << "]. Please raise an SPR."
	     << Sas::error("InvalidResponseChannels");
    }

  // Set up output file
  errstr <<  "Opening data set ..."
	 << Sas::message(Msg::App,Msg::Sparse);

  DataSet* set = dataSetServer->open( stringParameter("rmfset"), 
				      DataSet::Create);
  errstr <<  "Opened data set."
	 << Sas::message(Msg::App,Msg::Sparse);


  // Add rmftable
  errstr <<  "Adding MATRIX table. There will be "
	 << rmfDataServer->response()->numberOfEbins()
	 << " rows in table "
	 << Sas::message(Msg::App,Msg::Sparse);
  Table* rmftable = set->addTable( "MATRIX", 
				   rmfDataServer->response()->numberOfEbins()); 


  // Add attributes
  addAttributesTo( rmftable, 
		   "RSP_MATRIX");

  // Handle threshold value. Should ideally be placed in DataServer.
  double threshold = realParameter   ("threshold");

  *rmftable->addAttribute("LO_THRES",    
			  "Lower threshold used to construct the matrix")   
    = threshold;

  // Populate table by subtable iteration
  errstr <<  "Initialising table iterator .."
	 << Sas::message(Msg::App,Msg::Sparse);
  RmfTableIterator * rmftableiterator = 
    new RmfTableIterator( rmfDataServer, 
			  threshold );


  errstr <<  "Populating table .."
	 << Sas::message(Msg::App,Msg::Sparse);
  rmftableiterator->processEntireTable(rmftable); 

  /* Clean up table */
  delete rmftableiterator; rmftableiterator=0;

  /* Close and open the RMF to evade a bug which occurs in LOW_MEM mode */
  dataSetServer->close(set);

  set = dataSetServer->open( stringParameter("rmfset"), 
				      DataSet::Modify);

  //------------------------------

  // Add ebounds extension 

  //------------------------------

  errstr << "Adding EBOUNDS extension ... "
	 << Sas::message(Msg::App, Msg::Sparse);
  
  Table* ebounds_table = set->addTable( "EBOUNDS", 
					rmfDataServer->
					response()->
					numberOfChannels());


  /* Add attributes */

  addAttributesTo( ebounds_table, 
		   "EBOUNDS");

  /* Add CHANNEL, E_MIN and E_MAX columns */

  Column* channelCol = ebounds_table->addColumn("CHANNEL", Column::Int16, "", "");

  ColumnData* channelColData = channelCol->data();

  int16* channelData = channelColData->int16Data();

 /* Add min/max values foir this column */
  *(channelCol->addAttribute("TLMIN", "Minimum value legally allowed in FCHAN column")) = (int)tlmin;
                                                                                
  *(channelCol->addAttribute("TLMAX", "Minimum value legally allowed in FCHAN column")) = (int)tlmax;
                                                                                
  Column* e_minCol = ebounds_table->addColumn("E_MIN", Column::Real32, "", "keV");

  ColumnData* e_minColData = e_minCol->data();

  real32* e_minData = e_minColData->real32Data();


  Column* e_maxCol = ebounds_table->addColumn("E_MAX", Column::Real32, "", "keV");

  ColumnData* e_maxColData = e_maxCol->data();

  real32* e_maxData = e_maxColData->real32Data();

  
  /* Populate columns */

  const BinnedAxis * ebounds = rmfDataServer->response()->ebounds();

  if (ebounds_table->rows() <= ebounds->bins())
    {
      for (unsigned row = 0; row < ebounds_table->rows(); ++row)
	{
	  channelData[row] = row;
	  /* NOTE, response->ebounds returns ranges in eV 
	     - need to convert to keV */
	  e_minData[row] = ebounds->bin(row).lo()/1000.;
	  e_maxData[row] = ebounds->bin(row).hi()/1000.;
	}
    }
  else
    {
      errstr << " Requested size of ebounds table "
	     << ebounds_table->rows()
	     << " exceeds the data available. Will pad out the remaining rows with the highest ebounds pair available."
	     << Sas::message(Msg::App,Msg::Sparse);
 
      for (unsigned row = 0; row < ebounds->bins(); ++row)
	{
	  channelData[row] = row;
	  /* NOTE, response->ebounds returns ranges in eV 
	     - need to convert to keV */

	  e_minData[row] = ebounds->bin(row).lo()/1000.;
	  e_maxData[row] = ebounds->bin(row).hi()/1000.;

   	}

      unsigned lastbin = ebounds->bins()-1;
      double eminlast  = ebounds->bin(lastbin).lo()/1000.;
      double emaxlast  = ebounds->bin(lastbin).hi()/1000.;

      for (unsigned row = lastbin+1; row < ebounds_table->rows(); ++row)
	{
	  channelData[row] = row;
	  e_minData[row] = eminlast;
	  e_maxData[row] = emaxlast;
	}
      
    }
  
  /* Close dataset */
  /* Clean up pointers */
  errstr << " Cleanup ... "
	 << Sas::message(Msg::App,Msg::Sparse);

  delete channelColData;
  delete e_minColData;
  delete e_maxColData;

  delete rmfDataServer; 
  errstr << " Closing data set ... "
	 << Sas::message(Msg::App,Msg::Sparse);
  dataSetServer->close(set);
  errstr << " Closed data set."
	 << Sas::message(Msg::App,Msg::Sparse);


  errstr << " Cleanup complete. "
	 << Sas::message(Msg::App,Msg::Sparse);

  ENDPROC;

}

void Rmfgen::addAttributesTo(Table* table, 
			     const string& hduclas2)
{
  PROC;

      *table->addAttribute("TELESCOP",    "mission/satellite name")                  = _telescope;

      *table->addAttribute("INSTRUME",    "instrument/detector name")                = _instrument;

      *table->addAttribute("FILTER",      "filter in use")                           = _filter;

      *table->addAttribute("RMFVERSN",    "OGIP classification of FITS format")      = "1992a";

      *table->addAttribute("CHANTYPE",    "Detector Channel Type in use (PHA or PI") = "PI";

      *table->addAttribute("DETCHANS",    "total number of detector channels")       = (int)_detchans;
 
      *table->addAttribute("HDUCLASS",    "Format conforms to OGIP standard")        = "OGIP";

      *table->addAttribute("HDUCLAS1",    "dataset relates to spectral response")    = "RESPONSE";

      *table->addAttribute("HDUCLAS2",    "dataset is a spectral response matrix")   = hduclas2;

      *table->addAttribute("HDUCLAS3",    "photon redistribution matrix")            = "REDIST";

      *table->addAttribute("HDUVERS",    "Version of format (OGIP memo CAL/GEN/92-002a)")            = "1.3.0";

      *table->addAttribute("HDUVERS1",    "Obsolete - included for backwards compatibility")            = "1.0.0";

      *table->addAttribute("HDUVERS1",    "Obsolete - included for backwards compatibility")            = "1.3.0";

      ENDPROC;
}

